import json

from platone.module import Module
from platone.utils.builtin_abis import CNS_ABI
from platone.utils.transactions import send_obj_transaction


class CNS(Module):
    ABI = CNS_ABI
    is_wait_receipt = False

    def __init__(self, web3):
        super().__init__(web3)
        self.contract = self.web3.platone.contract(address=self.web3.cns_address, abi=self.ABI,
                                           ledger=self.web3.system_ledger)

    def getAllContracts(self):
        return self.contract.functions.GetAllContracts().call()

    def getContractAddress(self, name, version):
        return self.contract.functions.GetContractAddress(name, version).call()

    def getContractsByAddress(self, address):
        return self.contract.functions.GetContractsByAddr(address).call()

    def getContractsByOwner(self, address):
        return self.contract.functions.GetContractsByOwner(address).call()

    def Register(self, name, version, address, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.Register(name, version, address).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)

    def Unregister(self, name, version, private_key, tx_cfg: dict = None):
        txn = self.contract.functions.Unregister(name, version).buildTransaction(tx_cfg)
        return send_obj_transaction(self, txn, private_key, ledger=self.web3.system_ledger)



import json
import sha3
import rlp
import copy
from platone.packages.eth_utils.hexadecimal import remove_0x_prefix
from platone.packages.platone_account import (
    Account,
)
from platone.packages.platone_account.internal.transactions import bech32_address_bytes
from platone.packages.eth_utils import (
    apply_to_return_value,
    is_checksum_address,
    is_string,
)
from hexbytes import (
    HexBytes,
)

from platone.contract import (
    Contract,
)
from platone.module import (
    Module,
)
from platone.utils.blocks import (
    select_method_for_block_identifier,
)
from platone.utils.decorators import (
    deprecated_for,
)
from platone.utils.empty import (
    empty,
)
from platone.utils.encoding import (
    to_hex,
)
from platone.utils.filters import (
    BlockFilter,
    LogFilter,
    TransactionFilter,
)
from platone.utils.toolz import (
    assoc,
    merge,
)
from platone.utils.transactions import (
    assert_valid_transaction_params,
    extract_valid_transaction_params,
    get_buffered_gas_estimate,
    get_required_transaction,
    replace_transaction,
    wait_for_transaction_receipt,
)

from platone.packages.platone_account.internal.signing import (
    to_standard_signature_bytes,
)

true = True
false = False


class Platone(Module):
    account = Account()
    defaultAccount = empty
    defaultBlock = "latest"
    defaultContractFactory = Contract
    gasPriceStrategy = None

    @deprecated_for("doing nothing at all")
    def enable_unaudited_features(self):
        pass

    def namereg(self):
        raise NotImplementedError()

    def icapNamereg(self):
        raise NotImplementedError()

    @property
    def protocolVersion(self):
        return self.web3.manager.request_blocking("platon_protocolVersion", [])

    def syncing(self, ledger=None):
        return self.web3.manager.request_blocking("platon_syncing", self.set_ledger([], ledger))

    def gasPrice(self, ledger=None):
        return self.web3.manager.request_blocking("platon_gasPrice", self.set_ledger([], ledger))

    def accounts(self, ledger=None):
        return self.web3.manager.request_blocking("platon_accounts", self.set_ledger([], ledger))

    def blockNumber(self, ledger=None):
        return self.web3.manager.request_blocking("platon_blockNumber", self.set_ledger([], ledger))

    @property
    def evidences(self, ledger=None):
        data = self.web3.manager.request_blocking("platon_evidences", self.set_ledger([], ledger))
        return json.loads(data)

    @property
    def consensusStatus(self, ledger=None):
        return self.web3.manager.request_blocking("platon_consensusStatus", self.set_ledger([], ledger))

    def getPrepareQC(self, block_number, ledger=None):
        return self.web3.manager.request_blocking("platon_getPrepareQC", self.set_ledger([block_number], ledger))

    def getBalance(self, account, block_identifier=None, ledger=None):
        if block_identifier is None:
            block_identifier = self.defaultBlock
        return self.web3.manager.request_blocking(
            "platon_getBalance",
            self.set_ledger([account, block_identifier], ledger),
        )

    def getStorageAt(self, account, position, block_identifier=None, ledger=None):
        if block_identifier is None:
            block_identifier = self.defaultBlock
        return self.web3.manager.request_blocking(
            "platon_getStorageAt",
            self.set_ledger([account, position, block_identifier], ledger)
        )

    def getCode(self, account, block_identifier=None, ledger=None):
        if block_identifier is None:
            block_identifier = self.defaultBlock
        return self.web3.manager.request_blocking(
            "platon_getCode",
            self.set_ledger([account, block_identifier], ledger),
        )

    def getBlock(self, block_identifier, full_transactions=False, ledger=None):
        """
        `platon_getBlockByHash`
        `platon_getBlockByNumber`
        """
        method = select_method_for_block_identifier(
            block_identifier,
            if_predefined='platon_getBlockByNumber',
            if_hash='platon_getBlockByHash',
            if_number='platon_getBlockByNumber',
        )

        return self.web3.manager.request_blocking(
            method,
            self.set_ledger([block_identifier, full_transactions], ledger),
        )

    def getBlockTransactionCount(self, block_identifier, ledger=None):
        """
        `platon_getBlockTransactionCountByHash`
        `platon_getBlockTransactionCountByNumber`
        """
        method = select_method_for_block_identifier(
            block_identifier,
            if_predefined='platon_getBlockTransactionCountByNumber',
            if_hash='platon_getBlockTransactionCountByHash',
            if_number='platon_getBlockTransactionCountByNumber',
        )
        return self.web3.manager.request_blocking(
            method,
            self.set_ledger([block_identifier], ledger),
        )

    def getTransaction(self, transaction_hash, ledger=None):
        return self.web3.manager.request_blocking(
            "platon_getTransactionByHash",
            self.set_ledger([transaction_hash], ledger),
        )

    def getRawTransaction(self, transaction_hash, ledger=None):
        return self.web3.manager.request_blocking(
            "platon_getRawTransactionByHash",
            self.set_ledger([transaction_hash], ledger),
        )

    @deprecated_for("w3.platone.getTransactionByBlock")
    def getTransactionFromBlock(self, block_identifier, transaction_index, ledger=None):
        """
        Alias for the method getTransactionByBlock
        Depreceated to maintain naming consistency with the json-rpc API
        """
        return self.getTransactionByBlock(block_identifier, transaction_index, ledger)

    def getTransactionByBlock(self, block_identifier, transaction_index, ledger=None):
        """
        `platon_getTransactionByBlockHashAndIndex`
        `platon_getTransactionByBlockNumberAndIndex`
        """
        method = select_method_for_block_identifier(
            block_identifier,
            if_predefined='platon_getTransactionByBlockNumberAndIndex',
            if_hash='platon_getTransactionByBlockHashAndIndex',
            if_number='platon_getTransactionByBlockNumberAndIndex',
        )
        return self.web3.manager.request_blocking(
            method,
            self.set_ledger([block_identifier, transaction_index], ledger),
        )

    def waitForTransactionReceipt(self, transaction_hash, ledger=None, timeout=120):
        return wait_for_transaction_receipt(self.web3, transaction_hash, ledger, timeout)

    def getTransactionReceipt(self, transaction_hash, ledger=None):
        return self.web3.manager.request_blocking(
            "platon_getTransactionReceipt",
            self.set_ledger([transaction_hash], ledger),
        )

    def getTransactionCount(self, account, block_identifier=None, ledger=None):
        if block_identifier is None:
            block_identifier = self.defaultBlock
        return self.web3.manager.request_blocking(
            "platon_getTransactionCount",
            self.set_ledger([account, block_identifier], ledger),
        )

    def replaceTransaction(self, transaction_hash, new_transaction, ledger=None):
        if ledger is None:
            ledger = self.web3.ledgers
        current_transaction = get_required_transaction(self.web3, transaction_hash, ledger)
        return replace_transaction(self.web3, current_transaction, new_transaction, ledger)

    def modifyTransaction(self, transaction_hash, ledger=None, **transaction_params):
        assert_valid_transaction_params(transaction_params)
        current_transaction = get_required_transaction(self.web3, transaction_hash, ledger)
        current_transaction_params = extract_valid_transaction_params(current_transaction)
        new_transaction = merge(current_transaction_params, transaction_params)
        return replace_transaction(self.web3, current_transaction, new_transaction, ledger)

    def sendTransaction(self, transaction, ledger=None):
        # TODO: move to middleware
        if 'from' not in transaction and is_checksum_address(self.defaultAccount):
            transaction = assoc(transaction, 'from', self.defaultAccount)

        # TODO: move gas estimation in middleware
        if 'gas' not in transaction:
            transaction = assoc(
                transaction,
                'gas',
                get_buffered_gas_estimate(self.web3, transaction, ledger),
            )

        return self.web3.manager.request_blocking(
            "platon_sendTransaction",
            [transaction],
        )

    def sendRawTransaction(self, raw_transaction, ledger=None):
        return self.web3.manager.request_blocking(
            "platon_sendRawTransaction",
            self.set_ledger([raw_transaction], ledger),
        )

    def sign(self, account, data=None, hexstr=None, text=None):
        message_hex = to_hex(data, hexstr=hexstr, text=text)
        return self.web3.manager.request_blocking(
            "platon_sign", [account, message_hex],
        )

    @apply_to_return_value(HexBytes)
    def call(self, transaction, block_identifier=None, ledger=None):
        # TODO: move to middleware
        if 'from' not in transaction and is_checksum_address(self.defaultAccount):
            transaction = assoc(transaction, 'from', self.defaultAccount)

        # TODO: move to middleware
        if block_identifier is None:
            block_identifier = self.defaultBlock
        return self.web3.manager.request_blocking(
            "platon_call",
            self.set_ledger([transaction, block_identifier], ledger),
        )

    def estimateGas(self, transaction, ledger=None):
        # TODO: move to middleware
        if 'from' not in transaction and is_checksum_address(self.defaultAccount):
            transaction = assoc(transaction, 'from', self.defaultAccount)

        return self.web3.manager.request_blocking(
            "platon_estimateGas",
            self.set_ledger([transaction], ledger),
        )

    def filter(self, filter_params=None, filter_id=None, ledger=None):
        if filter_id and filter_params:
            raise TypeError(
                "Ambiguous invocation: provide either a `filter_params` or a `filter_id` argument. "
                "Both were supplied."
            )
        if is_string(filter_params):
            if filter_params == "latest":
                filter_id = self.web3.manager.request_blocking(
                    "platon_newBlockFilter", self.set_ledger([], ledger),
                )
                return BlockFilter(self.web3, filter_id)
            elif filter_params == "pending":
                filter_id = self.web3.manager.request_blocking(
                    "platon_newPendingTransactionFilter", self.set_ledger([], ledger),
                )
                return TransactionFilter(self.web3, filter_id)
            else:
                raise ValueError(
                    "The filter API only accepts the values of `pending` or "
                    "`latest` for string based filters"
                )
        elif isinstance(filter_params, dict):
            _filter_id = self.web3.manager.request_blocking(
                "platon_newFilter",
                self.set_ledger([filter_params], ledger),
            )
            return LogFilter(self.web3, _filter_id)
        elif filter_id and not filter_params:
            return LogFilter(self.web3, filter_id)
        else:
            raise TypeError("Must provide either filter_params as a string or "
                            "a valid filter object, or a filter_id as a string "
                            "or hex.")

    def getFilterChanges(self, filter_id, ledger=None):
        return self.web3.manager.request_blocking(
            "platon_getFilterChanges", self.set_ledger([filter_id], ledger),
        )

    def getFilterLogs(self, filter_id, ledger=None):
        return self.web3.manager.request_blocking(
            "platon_getFilterLogs", self.set_ledger([filter_id], ledger),
        )

    def getLogs(self, filter_params, ledger=None):
        return self.web3.manager.request_blocking(
            "platon_getLogs", self.set_ledger([filter_params], ledger),
        )

    def uninstallFilter(self, filter_id, ledger=None):
        return self.web3.manager.request_blocking(
            "platon_uninstallFilter", self.set_ledger([filter_id], ledger),
        )

    def wasm_type(self, abi_data):
        for i in range(len(abi_data)):
            if abi_data[i]['type'] == 'Action':
                abi_data[i]['type'] = 'function'
            if abi_data[i]['type'] == 'Event':
                abi_data[i]['type'] = 'event'
                abi_data[i]['anonymous'] = False
                if len(abi_data[i]['input']) > 0:
                    for j in range(len(abi_data[i]['input'])):
                        abi_data[i]['input'][j]['indexed'] = ((j + 1) <= abi_data[i]['topic'])
            if abi_data[i]['type'] == 'struct':
                if 'fields' in abi_data[i] and 'inputs' not in abi_data[i]:
                    abi_data[i]['inputs'] = abi_data[i].pop('fields')
                    if len(abi_data[i]['baseclass']) > 0:
                        for j in range(len(abi_data[i]['baseclass'])):
                            abi_data[i]['inputs'].insert(j, {'name': abi_data[i]['baseclass'][j], 'type': 'struct'})
                    # else :
                    #     abi_data[i]['inputs'].insert(0, {'name': abi_data[i]['baseclass'], 'type': 'struct'})
                    del abi_data[i]['baseclass']
            if abi_data[i]['name'] == 'init':
                abi_data[i]['type'] = 'constructor'
            if 'input' in abi_data[i]:
                abi_data[i]['inputs'] = abi_data[i].pop('input')
            if 'output' in abi_data[i]:
                abi_data[i]['outputs'] = {'name': "", 'type': abi_data[i]['output']}
                del abi_data[i]['output']
        return abi_data

    def contract(self, address=None, **kwargs):
        ContractFactoryClass = kwargs.pop('ContractFactoryClass', self.defaultContractFactory)
        ContractFactory = ContractFactoryClass.factory(self.web3, **kwargs)

        if address:
            return ContractFactory(address)
        else:
            return ContractFactory

    def wasmcontract(self,
                     address=None,
                     **kwargs):
        if 'vmtype' in kwargs:
            if kwargs['vmtype'] == 1:
                abi_data = copy.deepcopy(kwargs['abi'])
                kwargs['abi'] = self.wasm_type(abi_data)
            # del kwargs['vmtype']
        ContractFactoryClass = kwargs.pop('ContractFactoryClass', self.defaultContractFactory)
        # 若kwargs中有'ContractFactoryClass'这个key，则返回对应的value值，若无这个key，则返回self.defaultContractFactory
        ContractFactory = ContractFactoryClass.factory(self.web3, **kwargs)

        if address:
            return ContractFactory(address)
        else:
            return ContractFactory

    def setContractFactory(self, contractFactory):
        self.defaultContractFactory = contractFactory

    # @deprecated_in_v5
    # def getCompilers(self):
    #     return self.web3.manager.request_blocking("platon_getCompilers", [])

    # def getWork(self):
    #     return self.web3.manager.request_blocking("platon_getWork", [])

    def generateGasPrice(self, transaction_params=None):
        if self.gasPriceStrategy:
            return self.gasPriceStrategy(self.web3, transaction_params)

    def setGasPriceStrategy(self, gas_price_strategy):
        self.gasPriceStrategy = gas_price_strategy

    # add to platon
    def analyzeReceiptByHash(self, tx_hash, ledger=None):
        receipt = self.waitForTransactionReceipt(tx_hash, ledger)
        return self.analyzeReceipt(receipt)

    def analyzeReceipt(self, transaction_receipt):
        return self.web3.analyzeReceipt(transaction_receipt)

    def ecrecover(self, block_identifier):
        block = self.getBlock(block_identifier)
        extra = block.proofOfAuthorityData[0:32]
        sign = block.proofOfAuthorityData[32:]
        miner = bech32_address_bytes(remove_0x_prefix(block.miner))
        raw_data = [bytes.fromhex(remove_0x_prefix(block.parentHash.hex())),
                    miner,
                    bytes.fromhex(remove_0x_prefix(block.stateRoot.hex())),
                    bytes.fromhex(remove_0x_prefix(block.transactionsRoot.hex())),
                    bytes.fromhex(remove_0x_prefix(block.receiptsRoot.hex())),
                    bytes.fromhex(remove_0x_prefix(block.logsBloom.hex())),
                    block.number,
                    block.gasLimit,
                    block.gasUsed,
                    block.timestamp,
                    extra,
                    bytes.fromhex(remove_0x_prefix(block.nonce))
                    ]
        message_hash = sha3.keccak_256(rlp.encode(raw_data)).digest()
        hash_bytes = HexBytes(message_hash)
        signature_bytes = HexBytes(sign)
        signature_bytes_standard = to_standard_signature_bytes(signature_bytes)
        signature_obj = self.account._keys.Signature(signature_bytes=signature_bytes_standard)
        return remove_0x_prefix(signature_obj.recover_public_key_from_msg_hash(hash_bytes).to_hex())



from platone.module import (
    Module,
)


class TxPool(Module):
    def content(self, ledger=None):
        return self.web3.manager.request_blocking("txpool_content", self.set_ledger([], ledger=ledger))

    def inspect(self, ledger=None):
        return self.web3.manager.request_blocking("txpool_inspect", self.set_ledger([], ledger=ledger))

    def status(self, ledger=None):
        return self.web3.manager.request_blocking("txpool_status", self.set_ledger([], ledger=ledger))



from platone.module import (
    Module,
)


class Personal(Module):
    """
    https://github.com/ethereum/go-ethereum/wiki/Management-APIs#personal
    """
    def importRawKey(self, private_key, passphrase):
        return self.web3.manager.request_blocking(
            "personal_importRawKey",
            [private_key, passphrase],
        )

    def newAccount(self, password):
        return self.web3.manager.request_blocking(
            "personal_newAccount", [password],
        )

    @property
    def listAccounts(self):
        return self.web3.manager.request_blocking(
            "personal_listAccounts", [],
        )

    @property
    def listWallets(self):
        return self.web3.manager.request_blocking(
            "personal_listWallets", [],
        )

    def sendTransaction(self, transaction, passphrase, ledger=None):
        return self.web3.manager.request_blocking(
            "personal_sendTransaction",
            self.set_ledger([transaction, passphrase], ledger),
        )

    def lockAccount(self, account):
        return self.web3.manager.request_blocking(
            "personal_lockAccount",
            [account],
        )

    def unlockAccount(self, account, passphrase, duration=None):
        try:
            return self.web3.manager.request_blocking(
                "personal_unlockAccount",
                [account, passphrase, duration],
            )
        except ValueError as err:
            if "could not decrypt" in str(err):
                # Hack to handle go-ethereum error response.
                return False
            else:
                raise

    def sign(self, message, signer, passphrase, ledger=None):
        return self.web3.manager.request_blocking(
            'personal_sign',
            self.set_ledger([message, signer, passphrase], ledger),
        )

    def signTransaction(self, transaction_dict, passphrase, ledger=None):
        return self.web3.manager.request_blocking(
            'personal_signTransaction',
            self.set_ledger([transaction_dict, passphrase], ledger),
        )

    def ecRecover(self, message, signature):
        return self.web3.manager.request_blocking(
            'personal_ecRecover',
            [message, signature],
        )

    def openWallet(self, url, passphrase):
        return self.web3.manager.request_blocking(
            'personal_openWallet',
            [url, passphrase],
        )



import collections
import math
import operator

from platone.packages.eth_utils import (
    to_tuple,
)

from platone.exceptions import (
    InsufficientData,
    ValidationError,
)
from platone.utils.math import (
    percentile,
)
from platone.utils.toolz import (
    curry,
    groupby,
    sliding_window,
)

MinerData = collections.namedtuple(
    'MinerData',
    ['miner', 'num_blocks', 'min_gas_price', 'low_percentile_gas_price'])

Probability = collections.namedtuple('Probability', ['gas_price', 'prob'])


def _get_avg_block_time(w3, sample_size):
    latest = w3.platone.getBlock('latest')

    constrained_sample_size = min(sample_size, latest['number'])
    if constrained_sample_size == 0:
        raise ValidationError('Constrained sample size is 0')

    oldest = w3.platone.getBlock(latest['number'] - constrained_sample_size)
    return (latest['timestamp'] - oldest['timestamp']) / constrained_sample_size


def _get_raw_miner_data(w3, sample_size):
    latest = w3.platone.getBlock('latest', full_transactions=True)

    for transaction in latest['transactions']:
        yield (latest['miner'], latest['hash'], transaction['gasPrice'])

    block = latest

    for _ in range(sample_size - 1):
        if block['number'] == 0:
            break

        # we intentionally trace backwards using parent hashes rather than
        # block numbers to make caching the data easier to implement.
        block = w3.platone.getBlock(block['parentHash'], full_transactions=True)
        for transaction in block['transactions']:
            yield (block['miner'], block['hash'], transaction['gasPrice'])


def _aggregate_miner_data(raw_data):
    data_by_miner = groupby(0, raw_data)

    for miner, miner_data in data_by_miner.items():
        _, block_hashes, gas_prices = map(set, zip(*miner_data))
        try:
            price_percentile = percentile(gas_prices, percentile=20)
        except InsufficientData:
            price_percentile = min(gas_prices)
        yield MinerData(
            miner,
            len(set(block_hashes)),
            min(gas_prices),
            price_percentile)


@to_tuple
def _compute_probabilities(miner_data, wait_blocks, sample_size):
    """
    Computes the probabilities that a txn will be accepted at each of the gas
    prices accepted by the miners.
    """
    miner_data_by_price = tuple(sorted(
        miner_data,
        key=operator.attrgetter('low_percentile_gas_price'),
        reverse=True,
    ))
    for idx in range(len(miner_data_by_price)):
        low_percentile_gas_price = miner_data_by_price[idx].low_percentile_gas_price
        num_blocks_accepting_price = sum(m.num_blocks for m in miner_data_by_price[idx:])
        inv_prob_per_block = (sample_size - num_blocks_accepting_price) / sample_size
        probability_accepted = 1 - inv_prob_per_block ** wait_blocks
        yield Probability(low_percentile_gas_price, probability_accepted)


def _compute_gas_price(probabilities, desired_probability):
    """
    Given a sorted range of ``Probability`` named-tuples returns a gas price
    computed based on where the ``desired_probability`` would fall within the
    range.

    :param probabilities: An iterable of `Probability` named-tuples sorted in reverse order.
    :param desired_probability: An floating point representation of the desired
        probability. (e.g. ``85% -> 0.85``)
    """
    first = probabilities[0]
    last = probabilities[-1]

    if desired_probability >= first.prob:
        return int(first.gas_price)
    elif desired_probability <= last.prob:
        return int(last.gas_price)

    for left, right in sliding_window(2, probabilities):
        if desired_probability < right.prob:
            continue
        elif desired_probability > left.prob:
            # This code block should never be reachable as it would indicate
            # that we already passed by the probability window in which our
            # `desired_probability` is located.
            raise Exception('Invariant')

        adj_prob = desired_probability - right.prob
        window_size = left.prob - right.prob
        position = adj_prob / window_size
        gas_window_size = left.gas_price - right.gas_price
        gas_price = int(math.ceil(right.gas_price + gas_window_size * position))
        return gas_price
    else:
        # The initial `if/else` clause in this function handles the case where
        # the `desired_probability` is either above or below the min/max
        # probability found in the `probabilities`.
        #
        # With these two cases handled, the only way this code block should be
        # reachable would be if the `probabilities` were not sorted correctly.
        # Otherwise, the `desired_probability` **must** fall between two of the
        # values in the `probabilities``.
        raise Exception('Invariant')


@curry
def construct_time_based_gas_price_strategy(max_wait_seconds,
                                            sample_size=120,
                                            probability=98):
    """
    A gas pricing strategy that uses recently mined block data to derive a gas
    price for which a transaction is likely to be mined within X seconds with
    probability P.

    :param max_wait_seconds: The desired maxiumum number of seconds the
        transaction should take to mine.
    :param sample_size: The number of recent blocks to sample
    :param probability: An integer representation of the desired probability
        that the transaction will be mined within ``max_wait_seconds``.  0 means 0%
        and 100 means 100%.
    """
    def time_based_gas_price_strategy(web3, transaction_params):
        avg_block_time = _get_avg_block_time(web3, sample_size=sample_size)
        wait_blocks = int(math.ceil(max_wait_seconds / avg_block_time))

        raw_miner_data = _get_raw_miner_data(web3, sample_size=sample_size)
        miner_data = _aggregate_miner_data(raw_miner_data)

        probabilities = _compute_probabilities(
            miner_data,
            wait_blocks=wait_blocks,
            sample_size=sample_size,
        )

        gas_price = _compute_gas_price(probabilities, probability / 100)
        return gas_price
    return time_based_gas_price_strategy


# fast: mine within 1 minute
fast_gas_price_strategy = construct_time_based_gas_price_strategy(
    max_wait_seconds=60,
    sample_size=120,
)
# medium: mine within 10 minutes
medium_gas_price_strategy = construct_time_based_gas_price_strategy(
    max_wait_seconds=600,
    sample_size=120,
)
# slow: mine within 1 hour (60 minutes)
slow_gas_price_strategy = construct_time_based_gas_price_strategy(
    max_wait_seconds=60 * 60,
    sample_size=120,
)
# glacial: mine within the next 24 hours.
glacial_gas_price_strategy = construct_time_based_gas_price_strategy(
    max_wait_seconds=24 * 60 * 60,
    sample_size=720,
)



from platone.packages.eth_utils.curried import (
    apply_formatter_at_index,
    apply_formatter_if,
    apply_formatters_to_dict,
    is_null,
)
from hexbytes import (
    HexBytes,
)

from platone.exceptions import (
    ValidationError,
)
from platone.middleware.formatting import (
    construct_web3_formatting_middleware,
)
from platone.utils.toolz import (
    complement,
    compose,
    curry,
    dissoc,
)

MAX_EXTRADATA_LENGTH = 32


is_not_null = complement(is_null)


@curry
def validate_chain_id(web3, chain_id):
    if chain_id == web3.chainId:
        return chain_id
    else:
        raise ValidationError(
            "The transaction declared chain ID %r, "
            "but the connected node is on %r" % (
                chain_id,
                "UNKNOWN",
            )
        )


def check_extradata_length(val):
    if not isinstance(val, (str, int, bytes)):
        return val
    result = HexBytes(val)
    if len(result) > MAX_EXTRADATA_LENGTH:
        raise ValidationError(
            "The field extraData is %d bytes, but should be %d. "
            "It is quite likely that you are connected to a POA chain. "
            "Refer "
            "http://web3py.readthedocs.io/en/stable/middleware.html#geth-style-proof-of-authority "
            "for more details. The full extraData is: %r" % (
                len(result), MAX_EXTRADATA_LENGTH, result
            )
        )
    return val


def transaction_normalizer(transaction):
    return dissoc(transaction, 'chainId')


def transaction_param_validator(web3):
    transactions_params_validators = {
        'chainId': apply_formatter_if(
            # Bypass `validate_chain_id` if chainId can't be determined
            lambda _: is_not_null(web3.chainId),
            validate_chain_id(web3)
        ),
    }
    return apply_formatter_at_index(
        apply_formatters_to_dict(transactions_params_validators),
        1
    )


BLOCK_VALIDATORS = {
    'extraData': check_extradata_length,
}


block_validator = apply_formatter_if(
    is_not_null,
    apply_formatters_to_dict(BLOCK_VALIDATORS)
)


@curry
def chain_id_validator(web3):
    return compose(
        apply_formatter_at_index(transaction_normalizer, 1),
        transaction_param_validator(web3)
    )


def build_validators_with_web3(w3):
    return dict(
        request_formatters={
            'platon_sendTransaction': chain_id_validator(w3),
            'platon_estimateGas': chain_id_validator(w3),
            'platon_call': chain_id_validator(w3),
        },
        result_formatters={
            'platon_getBlockByHash': block_validator,
            'platon_getBlockByNumber': block_validator,
        },
    )


validation_middleware = construct_web3_formatting_middleware(build_validators_with_web3)



import os

from platone.packages.eth_utils import (
    ValidationError,
)

from platone.exceptions import (
    InfuraKeyNotFound,
)

INFURA_MAINNET_DOMAIN = 'mainnet.infura.io'
INFURA_ROPSTEN_DOMAIN = 'ropsten.infura.io'

WEBSOCKET_SCHEME = 'wss'
HTTP_SCHEME = 'https'


def load_api_key():
    # at web3py v5, drop old variable name INFURA_API_KEY
    key = os.environ.get(
        'WEB3_INFURA_API_KEY',
        os.environ.get('INFURA_API_KEY', '')
    )

    if key == '':
        raise InfuraKeyNotFound(
            "No Infura Project ID found. Please ensure "
            "that the environment variable WEB3_INFURA_API_KEY is set."
        )
    return key


def load_secret():
    secret = os.environ.get('WEB3_INFURA_API_SECRET', '')
    return secret


def build_http_headers():
    secret = load_secret()
    if secret:
        headers = {'auth': ('', secret)}
        return headers


def build_infura_url(domain):
    scheme = os.environ.get('WEB3_INFURA_SCHEME', WEBSOCKET_SCHEME)
    key = load_api_key()
    secret = load_secret()

    if scheme == WEBSOCKET_SCHEME:
        return "%s://:%s@%s/ws/v3/%s" % (scheme, secret, domain, key)
    elif scheme == HTTP_SCHEME:
        return "%s://%s/v3/%s" % (scheme, domain, key)
    else:
        raise ValidationError("Cannot connect to Infura with scheme %r" % scheme)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import (
    find_packages,
    setup,
)

extras_require = {
    'tester': [
        "eth-tester[py-evm]==0.1.0-beta.39",
        "py-geth>=2.0.1,<3.0.0",
    ],
    'testrpc': ["eth-testrpc>=1.3.3,<2.0.0"],
    'linter': [
        "flake8==3.4.1",
        "isort>=4.2.15,<4.3.5",
    ],
    'docs': [
        "mock",
        "sphinx-better-theme>=0.1.4",
        "click>=5.1",
        "configparser==3.5.0",
        "contextlib2>=0.5.4",
        # "eth-testrpc>=0.8.0",
        # "ethereum-tester-client>=1.1.0",
        "ethtoken",
        "py-geth>=1.4.0",
        "py-solc>=0.4.0",
        "pytest>=2.7.2",
        "sphinx",
        "sphinx_rtd_theme>=0.1.9",
        "toposort>=1.4",
        "urllib3",
        "web3>=2.1.0",
        "wheel"
    ],
    'dev': [
        "bumpversion",
        "flaky>=3.3.0",
        "hypothesis>=3.31.2",
        "pytest>=3.5.0,<4",
        "pytest-mock==1.*",
        "pytest-pythonpath>=0.3",
        "pytest-watch==4.*",
        "pytest-xdist==1.*",
        "tox>=1.8.0",
        "tqdm",
        "twine",
        "when-changed"
    ]
}

extras_require['dev'] = (
    extras_require['tester'] +
    extras_require['linter'] +
    extras_require['docs'] +
    extras_require['dev']
)

setup(
    name='platone.py',
    # *IMPORTANT*: Don't manually change the version here. Use the 'bumpversion' utility.
    version='3.3.0',
    description="""PlatONE python SDK """,
    # long_description_markdown_filename='README.md',
    author='Piper shinnng',
    author_email='shinnng@outlook.com',
    url='https://github.com/PlatONEnetwork/PlatONE-Python-SDK',
    include_package_data=True,
    install_requires=[
        "toolz>=0.9.0,<1.0.0;implementation_name=='pypy'",
        "cytoolz>=0.9.0,<1.0.0;implementation_name=='cpython'",
        "six>=1.15.0",
        "numpy>=1.19.0",
        "eth-utils>=1.2.0,<2.0.0",
        "hexbytes>=0.1.0,<1.0.0",
        "lru-dict>=1.1.6,<2.0.0",
        "eth-hash[pycryptodome]>=0.2.0,<1.0.0",
        "requests>=2.16.0,<3.0.0",
        "websockets>=6.0.0,<7.0.0",
        "pypiwin32>=223;platform_system=='Windows'", 'rlp', 'pysha3'
    ],
    setup_requires=['setuptools-markdown'],
    python_requires='>=3.6,<4',
    extras_require=extras_require,
    py_modules=['web3'],
    license="MIT",
    zip_safe=False,
    keywords='platone',
    packages=find_packages(exclude=["tests", "tests.*"]),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)



import functools
import threading
import warnings


class combomethod:
    def __init__(self, method):
        self.method = method

    def __get__(self, obj=None, objtype=None):
        @functools.wraps(self.method)
        def _wrapper(*args, **kwargs):
            if obj is not None:
                return self.method(obj, *args, **kwargs)
            else:
                return self.method(objtype, *args, **kwargs)
        return _wrapper


def reject_recursive_repeats(to_wrap):
    '''
    Prevent simple cycles by returning None when called recursively with same instance
    '''
    to_wrap.__already_called = {}

    @functools.wraps(to_wrap)
    def wrapped(*args):
        arg_instances = tuple(map(id, args))
        thread_id = threading.get_ident()
        thread_local_args = (thread_id,) + arg_instances
        if thread_local_args in to_wrap.__already_called:
            raise ValueError('Recursively called %s with %r' % (to_wrap, args))
        to_wrap.__already_called[thread_local_args] = True
        try:
            wrapped_val = to_wrap(*args)
        finally:
            del to_wrap.__already_called[thread_local_args]
        return wrapped_val
    return wrapped


def deprecated_for(replace_message):
    '''
    Decorate a deprecated function, with info about what to use instead, like:

    @deprecated_for("toBytes()")
    def toAscii(arg):
        ...
    '''
    def decorator(to_wrap):
        @functools.wraps(to_wrap)
        def wrapper(*args, **kwargs):
            warnings.warn(
                "%s is deprecated in favor of %s" % (to_wrap.__name__, replace_message),
                category=DeprecationWarning,
                stacklevel=2)
            return to_wrap(*args, **kwargs)
        return wrapper
    return decorator



from typing import Any  # noqa: F401

from platone.packages.platone_keys.datatypes import (
    BaseSignature,
    NonRecoverableSignature,
    PrivateKey,
    PublicKey,
    Signature,
)


class BaseECCBackend(object):
    def ecdsa_sign(self,
                   msg_hash: bytes,
                   private_key: PrivateKey) -> Signature:
        raise NotImplementedError()

    def ecdsa_sign_non_recoverable(self,
                                   msg_hash: bytes,
                                   private_key: PrivateKey) -> NonRecoverableSignature:
        raise NotImplementedError()

    def ecdsa_verify(self,
                     msg_hash: bytes,
                     signature: BaseSignature,
                     public_key: PublicKey) -> bool:
        raise NotImplementedError()

    def ecdsa_recover(self,
                      msg_hash: bytes,
                      signature: Signature) -> PublicKey:
        raise NotImplementedError()

    def private_key_to_public_key(self,
                                  private_key: PrivateKey) -> PublicKey:
        raise NotImplementedError()

    def decompress_public_key_bytes(self,
                                    compressed_public_key_bytes: bytes) -> bytes:
        raise NotImplementedError()

    def compress_public_key_bytes(self,
                                  uncompressed_public_key_bytes: bytes) -> bytes:
        raise NotImplementedError()



from random import choice

xor = lambda a, b:list(map(lambda x, y: x ^ y, a, b))

rotl = lambda x, n:((x << n) & 0xffffffff) | ((x >> (32 - n)) & 0xffffffff)

get_uint32_be = lambda key_data:((key_data[0] << 24) | (key_data[1] << 16) | (key_data[2] << 8) | (key_data[3]))

put_uint32_be = lambda n:[((n>>24)&0xff), ((n>>16)&0xff), ((n>>8)&0xff), ((n)&0xff)]

padding = lambda data, block=16: data + [(16 - len(data) % block)for _ in range(16 - len(data) % block)]

unpadding = lambda data: data[:-data[-1]]

list_to_bytes = lambda data: b''.join([bytes((i,)) for i in data])

bytes_to_list = lambda data: [i for i in data]

random_hex = lambda x: ''.join([choice('0123456789abcdef') for _ in range(x)])




from platone.utils.toolz import (
    assoc,
    curry,
    merge,
)


def construct_formatting_middleware(
        request_formatters=None,
        result_formatters=None,
        error_formatters=None):
    def ignore_web3_in_standard_formatters(w3):
        return dict(
            request_formatters=request_formatters or {},
            result_formatters=result_formatters or {},
            error_formatters=error_formatters or {},
        )

    return construct_web3_formatting_middleware(ignore_web3_in_standard_formatters)


def construct_web3_formatting_middleware(web3_formatters_builder):
    def formatter_middleware(make_request, w3):
        formatters = merge(
            {
                'request_formatters': {},
                'result_formatters': {},
                'error_formatters': {},
            },
            web3_formatters_builder(w3),
        )
        return apply_formatters(make_request=make_request, **formatters)

    return formatter_middleware


@curry
def apply_formatters(
        method,
        params,
        make_request,
        request_formatters,
        result_formatters,
        error_formatters):
    if method in request_formatters:
        formatter = request_formatters[method]
        formatted_params = formatter(params)
        response = make_request(method, formatted_params)
    else:
        response = make_request(method, params)

    if 'result' in response and method in result_formatters:
        if response['result'] == None:
            return response
        formatter = result_formatters[method]
        formatted_response = assoc(
            response,
            'result',
            formatter(response['result']),
        )
        return formatted_response
    elif 'error' in response and method in error_formatters:
        formatter = error_formatters[method]
        formatted_response = assoc(
            response,
            'error',
            formatter(response['error']),
        )
        return formatted_response
    else:
        return response



import sys
import warnings

import pkg_resources

from .abi import (  # noqa: F401
    event_abi_to_log_topic,
    event_signature_to_log_topic,
    function_abi_to_4byte_selector,
    function_signature_to_4byte_selector,
    topic_decode,
)
from .address import (  # noqa: F401
    is_address,
    is_binary_address,
    is_canonical_address,
    is_checksum_address,
    is_checksum_formatted_address,
    is_hex_address,
    is_normalized_address,
    is_same_address,
    to_canonical_address,
    to_checksum_address,
    to_normalized_address,
)
from .applicators import (  # noqa: F401
    apply_formatter_at_index,
    apply_formatter_if,
    apply_formatter_to_array,
    apply_formatters_to_dict,
    apply_formatters_to_sequence,
    apply_key_map,
    apply_one_of_formatters,
    combine_argument_formatters,
)
from .conversions import (  # noqa: F401
    hexstr_if_str,
    text_if_str,
    to_bytes,
    to_hex,
    to_int,
    to_text,
)
from .crypto import keccak  # noqa: F401
from .currency import denoms, from_wei, to_wei  # noqa: F401
from .decorators import combomethod, replace_exceptions  # noqa: F401
from .encoding import big_endian_to_int, int_to_big_endian  # noqa: F401
from .exceptions import ValidationError  # noqa: F401
from .functional import (  # noqa: F401
    apply_to_return_value,
    flatten_return,
    reversed_return,
    sort_return,
    to_dict,
    to_list,
    to_ordered_dict,
    to_set,
    to_tuple,
)
from .hexadecimal import (  # noqa: F401
    add_0x_prefix,
    decode_hex,
    encode_hex,
    is_0x_prefixed,
    is_hex,
    is_hexstr,
    remove_0x_prefix,
)
from .humanize import (  # noqa: F401
    humanize_bytes,
    humanize_hash,
    humanize_integer_sequence,
    humanize_ipfs_uri,
    humanize_seconds,
)
from .logging import (  # noqa: F401
    DEBUG2_LEVEL_NUM,
    ExtendedDebugLogger,
    HasExtendedDebugLogger,
    HasExtendedDebugLoggerMeta,
    HasLogger,
    HasLoggerMeta,
    get_extended_debug_logger,
    get_logger,
    setup_DEBUG2_logging,
)
from .module_loading import import_string  # noqa: F401
from .numeric import clamp  # noqa: F401
from .types import (  # noqa: F401
    is_boolean,
    is_bytes,
    is_dict,
    is_integer,
    is_list,
    is_list_like,
    is_null,
    is_number,
    is_string,
    is_text,
    is_tuple,
)

if sys.version_info.major < 3:
    warnings.simplefilter("always", DeprecationWarning)
    warnings.warn(
        DeprecationWarning(
            "The `eth-utils` library has dropped support for Python 2. Upgrade to Python 3."
        )
    )
    warnings.resetwarnings()


__version__ = pkg_resources.get_distribution("eth-utils").version



from abc import (
    ABC,
    abstractmethod,
)


class BaseAccount(ABC):
    '''
    Abstract class that defines a collection of convenience methods
    to sign transactions and message hashes.
    '''

    @property
    @abstractmethod
    def address(self):
        '''
        The checksummed public address for this account.

        .. code-block:: python

            >>> my_account.address
            "0xF0109fC8DF283027b6285cc889F5aA624EaC1F55"

        '''
        pass

    @abstractmethod
    def signHash(self, message_hash):
        '''
        Sign the hash of a message, as in :meth:`~platone_account.account.Account.signHash`
        but without specifying the private key.

        :var bytes message_hash: 32 byte hash of the message to sign
        '''
        pass

    @abstractmethod
    def signTransaction(self, transaction_dict):
        '''
        Sign a transaction, as in :meth:`~platone_account.account.Account.signTransaction`
        but without specifying the private key.

        :var dict transaction_dict: transaction with all fields specified
        '''
        pass

    def __eq__(self, other):
        '''
        Equality test between two accounts.

        Two accounts are considered the same if they are exactly the same type,
        and can sign for the same address.
        '''
        return type(self) == type(other) and self.address == other.address

    def __hash__(self):
        return hash((type(self), self.address))



import itertools

from platone.packages.eth_abi import (
    decode_abi,
    decode_single,
    encode_single,
)
from platone.packages.eth_abi.abi import (
    process_type,
)
from platone.packages.eth_utils import (
    encode_hex,
    event_abi_to_log_topic,
    is_list_like,
    to_tuple,
    topic_decode,
)

from platone.datastructures import (
    AttributeDict,
)
from platone.exceptions import (
    MismatchedABI,
)
from platone.utils.encoding import (
    hexstr_if_str,
    to_bytes,
)
from platone.utils.normalizers import (
    BASE_RETURN_NORMALIZERS,
)
from platone.utils.toolz import (
    compose,
    curry,
)

from .abi import (
    exclude_indexed_event_inputs,
    get_abi_input_names,
    get_indexed_event_inputs,
    map_abi_data,
    normalize_event_input_types,
)
from platone.utils.contracts import wasmevent_decode

def construct_event_topic_set(event_abi, arguments=None):
    if arguments is None:
        arguments = {}
    if isinstance(arguments, (list, tuple)):
        if len(arguments) != len(event_abi['inputs']):
            raise ValueError(
                "When passing an argument list, the number of arguments must "
                "match the event constructor."
            )
        arguments = {
            arg['name']: [arg_value]
            for arg, arg_value
            in zip(event_abi['inputs'], arguments)
        }

    normalized_args = {
        key: value if is_list_like(value) else [value]
        for key, value in arguments.items()
    }

    event_topic = encode_hex(event_abi_to_log_topic(event_abi))
    indexed_args = get_indexed_event_inputs(event_abi)
    zipped_abi_and_args = [
        (arg, normalized_args.get(arg['name'], [None]))
        for arg in indexed_args
    ]
    encoded_args = [
        [
            None if option is None else encode_hex(encode_single(arg['type'], option))
            for option in arg_options]
        for arg, arg_options in zipped_abi_and_args
    ]

    topics = list(normalize_topic_list([event_topic] + encoded_args))
    return topics


def construct_event_data_set(event_abi, arguments=None):
    if arguments is None:
        arguments = {}
    if isinstance(arguments, (list, tuple)):
        if len(arguments) != len(event_abi['inputs']):
            raise ValueError(
                "When passing an argument list, the number of arguments must "
                "match the event constructor."
            )
        arguments = {
            arg['name']: [arg_value]
            for arg, arg_value
            in zip(event_abi['inputs'], arguments)
        }

    normalized_args = {
        key: value if is_list_like(value) else [value]
        for key, value in arguments.items()
    }

    non_indexed_args = exclude_indexed_event_inputs(event_abi)
    zipped_abi_and_args = [
        (arg, normalized_args.get(arg['name'], [None]))
        for arg in non_indexed_args
    ]
    encoded_args = [
        [
            None if option is None else encode_hex(encode_single(arg['type'], option))
            for option in arg_options]
        for arg, arg_options in zipped_abi_and_args
    ]

    data = [
        list(permutation)
        if any(value is not None for value in permutation)
        else []
        for permutation in itertools.product(*encoded_args)
    ]
    return data


def is_dynamic_sized_type(_type):
    base_type, type_size, arrlist = process_type(_type)
    if arrlist:
        return True
    elif base_type == 'string':
        return True
    elif base_type == 'bytes' and type_size == '':
        return True
    return False


@to_tuple
def get_event_abi_types_for_decoding(event_inputs):
    """
    Event logs use the `sha3(value)` for indexed inputs of type `bytes` or
    `string`.  Because of this we need to modify the types so that we can
    decode the log entries using the correct types.
    """
    for input_abi in event_inputs:
        if input_abi['indexed'] and is_dynamic_sized_type(input_abi['type']):
            yield 'bytes32'
        else:
            yield input_abi['type']


def get_event_data(event_abi, log_entry, vmtype=None):
    """
    Given an event ABI and a log entry for that event, return the decoded
    event data
    """
    if event_abi['anonymous']:
        log_topics = log_entry['topics']
    elif not log_entry['topics']:
        raise MismatchedABI("Expected non-anonymous event to have 1 or more topics")
    elif event_abi_to_log_topic(event_abi,vmtype) != log_entry['topics'][0]:
        raise MismatchedABI("The event signature did not match the provided ABI")
    else:
        log_topics = log_entry['topics'][1:]

    log_topics_abi = get_indexed_event_inputs(event_abi)
    log_topic_normalized_inputs = normalize_event_input_types(log_topics_abi)
    log_topic_types = get_event_abi_types_for_decoding(log_topic_normalized_inputs)
    log_topic_names = get_abi_input_names({'inputs': log_topics_abi})

    if len(log_topics) != len(log_topic_types):
        raise ValueError("Expected {0} log topics.  Got {1}".format(
            len(log_topic_types),
            len(log_topics),
        ))

    log_data = hexstr_if_str(to_bytes, log_entry['data'])
    log_data_abi = exclude_indexed_event_inputs(event_abi)
    log_data_normalized_inputs = normalize_event_input_types(log_data_abi)
    log_data_types = get_event_abi_types_for_decoding(log_data_normalized_inputs)
    log_data_names = get_abi_input_names({'inputs': log_data_abi})

    # sanity check that there are not name intersections between the topic
    # names and the data argument names.
    duplicate_names = set(log_topic_names).intersection(log_data_names)
    if duplicate_names:
        raise ValueError(
            "Invalid Event ABI:  The following argument names are duplicated "
            "between event inputs: '{0}'".format(', '.join(duplicate_names))
        )
    decoded_topic_data = [
        decode_single(topic_type, topic_data)
        for topic_type, topic_data
        in zip(log_topic_types, log_topics)
    ]
    normalized_topic_data = map_abi_data(
        BASE_RETURN_NORMALIZERS,
        log_topic_types,
        decoded_topic_data
    )
    if vmtype:
        normalized_log_data = wasmevent_decode(log_data_types, log_data)
        if isinstance(normalized_topic_data[0],bytes):
            if normalized_topic_data[0][0]==0:
                normalized_topic_data = topic_decode(normalized_topic_data)
    else:
        decoded_log_data = decode_abi(log_data_types, log_data)
        normalized_log_data = map_abi_data(
            BASE_RETURN_NORMALIZERS,
            log_data_types,
            decoded_log_data
        )

    event_args = dict(itertools.chain(
        zip(log_topic_names, normalized_topic_data),
        zip(log_data_names, normalized_log_data),
    ))

    event_data = {
        'args': event_args,
        'event': event_abi['name'],
        'logIndex': log_entry['logIndex'],
        'transactionIndex': log_entry['transactionIndex'],
        'transactionHash': log_entry['transactionHash'],
        'address': log_entry['address'],
        'blockHash': log_entry['blockHash'],
        'blockNumber': log_entry['blockNumber'],
    }

    return AttributeDict.recursive(event_data)


@to_tuple
def pop_singlets(seq):
    yield from (i[0] if is_list_like(i) and len(i) == 1 else i for i in seq)


@curry
def remove_trailing_from_seq(seq, remove_value=None):
    index = len(seq)
    while index > 0 and seq[index - 1] == remove_value:
        index -= 1
    return seq[:index]


normalize_topic_list = compose(
    remove_trailing_from_seq(remove_value=None),
    pop_singlets,)



from platone.utils.normalizers import (
    abi_ens_resolver,
)
from platone.utils.rpc_abi import (
    RPC_ABIS,
    abi_request_formatters,
)

from .formatting import (
    construct_formatting_middleware,
)


def name_to_address_middleware(w3):
    normalizers = [
        abi_ens_resolver(w3),
    ]
    return construct_formatting_middleware(
        # request_formatters=abi_request_formatters(normalizers, RPC_ABIS)
    )



from platone.utils.toolz import (
    excepts,
)


def construct_exception_handler_middleware(method_handlers=None):
    if method_handlers is None:
        method_handlers = {}

    def exception_handler_middleware(make_request, web3):
        def middleware(method, params):
            if method in method_handlers:
                exc_type, handler = method_handlers[method]
                return excepts(
                    exc_type,
                    make_request,
                    handler,
                )(method, params)
            else:
                return make_request(method, params)
        return middleware
    return exception_handler_middleware



import os
from urllib.parse import (
    urlparse,
)

from platone.exceptions import (
    CannotHandleRequest,
)
from platone.providers import (
    BaseProvider,
    HTTPProvider,
    IPCProvider,
    WebsocketProvider,
)

HTTP_SCHEMES = {'http', 'https'}
WS_SCHEMES = {'ws', 'wss'}


def load_provider_from_environment():
    uri_string = os.environ.get('WEB3_PROVIDER_URI', '')
    if not uri_string:
        return None

    return load_provider_from_uri(uri_string)


def load_provider_from_uri(uri_string, headers=None):
    uri = urlparse(uri_string)
    if uri.scheme == 'file':
        return IPCProvider(uri.path)
    elif uri.scheme in HTTP_SCHEMES:
        return HTTPProvider(uri_string, headers)
    elif uri.scheme in WS_SCHEMES:
        return WebsocketProvider(uri_string)
    else:
        raise NotImplementedError(
            'Web3 does not know how to connect to scheme %r in %r' % (
                uri.scheme,
                uri_string,
            )
        )


class AutoProvider(BaseProvider):

    default_providers = (
        load_provider_from_environment,
        IPCProvider,
        HTTPProvider,
        WebsocketProvider,
    )
    _active_provider = None

    def __init__(self, potential_providers=None):
        '''
        :param iterable potential_providers: ordered series of provider classes to attempt with

        AutoProvider will initialize each potential provider (without arguments),
        in an attempt to find an active node. The list will default to
        :attribute:`default_providers`.
        '''
        if potential_providers:
            self._potential_providers = potential_providers
        else:
            self._potential_providers = self.default_providers

    def make_request(self, method, params):
        try:
            return self._proxy_request(method, params)
        except IOError as exc:
            return self._proxy_request(method, params, use_cache=False)

    def isConnected(self):
        provider = self._get_active_provider(use_cache=True)
        return provider is not None and provider.isConnected()

    def _proxy_request(self, method, params, use_cache=True):
        provider = self._get_active_provider(use_cache)
        if provider is None:
            raise CannotHandleRequest("Could not discover provider")

        return provider.make_request(method, params)

    def _get_active_provider(self, use_cache):
        if use_cache and self._active_provider is not None:
            return self._active_provider

        for Provider in self._potential_providers:
            provider = Provider()
            if provider is not None and provider.isConnected():
                self._active_provider = provider
                return provider

        return None



import pkg_resources

from platone.packages.eth_abi.abi import (
    decode_single,
    decode_abi,
    encode_single,
    encode_abi,
    is_encodable,
)
try:
   __version__ = pkg_resources.get_distribution('eth-abi').version
except:
    __version__='0.7.1'


from platone.packages.eth_utils import (
    apply_to_return_value,
    add_0x_prefix,
    from_wei,
    is_address,
    # is_checksum_address,
    keccak,
    remove_0x_prefix,
    # to_checksum_address,
    to_wei,
)
from platone.packages.platone_keys.utils.address import MIANNETHRP, TESTNETHRP

from platone.packages.ens import ENS

from platone.admin import Admin
from platone.platone import Platone
from platone.miner import Miner
from platone.net import Net
from platone.personal import Personal
from platone.txpool import TxPool
from platone.debug import Debug

from platone.providers.eth_tester import (
    EthereumTesterProvider,
)
from platone.providers.ipc import (
    IPCProvider,
)
from platone.providers.rpc import (
    HTTPProvider,
)
from platone.providers.tester import (
    TestRPCProvider,
)
from platone.providers.websocket import (
    WebsocketProvider
)

from platone.manager import (
    RequestManager,
)

from platone.utils.abi import (
    map_abi_data,
)
from hexbytes import (
    HexBytes,
)
from platone.utils.decorators import (
    combomethod,
)
from platone.utils.empty import empty
from platone.utils.encoding import (
    hex_encode_abi_type,
    to_bytes,
    to_int,
    to_hex,
    to_text,
    analyze,
)
from platone.utils.normalizers import (
    abi_ens_resolver,
)


def get_default_modules():
    return {
        "platone": Platone,
        "net": Net,
        "personal": Personal,
        "txpool": TxPool,
        "miner": Miner,
        "admin": Admin,
        "debug": Debug
    }


def default_address(mainnet, testnet):
    return {MIANNETHRP: mainnet, TESTNETHRP: testnet}


NODE_MANAGER = default_address("lat1zqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqp7pn3ep",
                               "lax1zqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqp3yp7hw")
USER_MANAGER = default_address("lat1zqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqzsjx8h7",
                               "lax1zqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqzlh5ge3")
PARAM_MANAGER = default_address("lat1zqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqrdyjj2v",
                                "lax1zqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqrzpqayr")
CNS = default_address("lat1zqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqyva9ztf", "lax1zqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqyrchd9x")
LEDGER_MANAGER = default_address("lat1zqqqqqqqqqqqqqqqqqqqqqqqqqqqqqq93t3hkm",
                                 "lax1zqqqqqqqqqqqqqqqqqqqqqqqqqqqqqq97wrcc5")
LEDGER_NODE_MANAGER = default_address("lat1zqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqxlcypcy",
                                      "lax1zqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqxsakwkt")
RE_ENC = default_address("lat1zqqqqqqqqqqqqqqqqqqqqqqqqqqqqqq8zws59k", "lax1zqqqqqqqqqqqqqqqqqqqqqqqqqqqqqq8dtzmte")


def to_checksum_address(val):
    return val


def is_checksum_address(val):
    return True


class Web3:
    # Providers
    HTTPProvider = HTTPProvider
    IPCProvider = IPCProvider
    TestRPCProvider = TestRPCProvider
    EthereumTesterProvider = EthereumTesterProvider
    WebsocketProvider = WebsocketProvider

    # Managers
    RequestManager = RequestManager

    # Encoding and Decoding
    toBytes = staticmethod(to_bytes)
    toInt = staticmethod(to_int)
    toHex = staticmethod(to_hex)
    toText = staticmethod(to_text)

    # Currency Utility
    toWei = staticmethod(to_wei)
    fromWei = staticmethod(from_wei)
    analyzeReceipt = staticmethod(analyze)

    # Address Utility
    isAddress = staticmethod(is_address)
    isChecksumAddress = staticmethod(is_checksum_address)
    toChecksumAddress = staticmethod(to_checksum_address)

    def __init__(self, providers=empty, middlewares=None, modules=None, ens=empty, chain_id=100,
                 multi_ledger: bool = False, encryption_mode='SM'):
        self.manager = RequestManager(self, providers, middlewares)

        if modules is None:
            modules = get_default_modules()
        for module_name, module_class in modules.items():
            module_class.attach(self, module_name)

        self.ens = ens

        self.chain_id = chain_id
        if chain_id == 100:
            self.net_type = MIANNETHRP
        else:
            self.net_type = TESTNETHRP

        self.system_ledger = None
        self.current_ledger = None
        self.ledgers = None
        if multi_ledger is True:
            self.system_ledger = 'sys'
            self.current_ledger = self.system_ledger
            self.ledgers = [self.system_ledger]

        self.encryption_mode = encryption_mode

        # built-in contract address
        self.node_manager_address = NODE_MANAGER[self.net_type]
        self.user_manager_address = USER_MANAGER[self.net_type]
        self.param_manager_address = PARAM_MANAGER[self.net_type]
        self.cns_address = CNS[self.net_type]
        self.ledger_manager_address = LEDGER_MANAGER[self.net_type]
        self.ledger_node_manager_address = LEDGER_NODE_MANAGER[self.net_type]
        self.re_enc_address = RE_ENC[self.net_type]

    @property
    def currentLedger(self):
        return self.current_ledger

    @currentLedger.setter
    def currentLedger(self, ledger):
        self.current_ledger = ledger
        if ledger not in self.ledgers:
            self.ledgers.append(ledger)

    @property
    def chainId(self):
        return self.chain_id

    @chainId.setter
    def chainId(self, chain_id):
        self.chain_id = chain_id

    @property
    def middleware_stack(self):
        return self.manager.middleware_stack

    @property
    def providers(self):
        return self.manager.providers

    @providers.setter
    def providers(self, providers):
        self.manager.providers = providers

    @staticmethod
    @apply_to_return_value(HexBytes)
    def sha3(primitive=None, text=None, hexstr=None):
        if isinstance(primitive, (bytes, int, type(None))):
            input_bytes = to_bytes(primitive, hexstr=hexstr, text=text)
            return keccak(input_bytes)

        raise TypeError(
            "You called sha3 with first arg %r and keywords %r. You must call it with one of "
            "these approaches: sha3(text='txt'), sha3(hexstr='0x747874'), "
            "sha3(b'\\x74\\x78\\x74'), or sha3(0x747874)." % (
                primitive,
                {'text': text, 'hexstr': hexstr}
            )
        )

    @combomethod
    def soliditySha3(cls, abi_types, values):
        """
        Executes sha3 (keccak256) exactly as Solidity does.
        Takes list of abi_types as inputs -- `[uint24, int8[], bool]`
        and list of corresponding values  -- `[20, [-1, 5, 0], True]`
        """
        if len(abi_types) != len(values):
            raise ValueError(
                "Length mismatch between provided abi types and values.  Got "
                "{0} types and {1} values.".format(len(abi_types), len(values))
            )

        if isinstance(cls, type):
            w3 = None
        else:
            w3 = cls
        normalized_values = map_abi_data([abi_ens_resolver(w3)], abi_types, values)

        hex_string = add_0x_prefix(''.join(
            remove_0x_prefix(hex_encode_abi_type(abi_type, value))
            for abi_type, value
            in zip(abi_types, normalized_values)
        ))
        return cls.sha3(hexstr=hex_string)

    def isConnected(self):
        for provider in self.providers:
            if provider.isConnected():
                return True
        else:
            return False

    @property
    def ens(self):
        if self._ens is empty:
            return ENS.fromWeb3(self)
        else:
            return self._ens

    @ens.setter
    def ens(self, new_ens):
        self._ens = new_ens

    def pubkey_to_address(self, pubkey):
        addr_dict = {MIANNETHRP: pubkey.to_bech32_address(),
                     TESTNETHRP: pubkey.to_bech32_test_address()}
        return addr_dict[self.net_type]



from platone.packages.eth_utils import (
    is_string,
)

from platone.utils.formatters import (
    apply_formatter_at_index,
    apply_formatter_if,
    apply_formatters_to_dict,
)

from .formatting import (
    construct_formatting_middleware,
)

FILTER_PARAM_NORMALIZERS = apply_formatters_to_dict({
    'address': apply_formatter_if(is_string, lambda x: [x])})

METHOD_NORMALIZERS = {
    'platon_getLogs': apply_formatter_at_index(FILTER_PARAM_NORMALIZERS, 1),
    'platon_newFilter': apply_formatter_at_index(FILTER_PARAM_NORMALIZERS, 1)
}

request_parameter_normalizer = construct_formatting_middleware(
    request_formatters=METHOD_NORMALIZERS,
)



from typing import Callable, TypeVar, Union

from eth_typing import HexStr, Primitives

from .decorators import validate_conversion_arguments
from .encoding import big_endian_to_int, int_to_big_endian
from .hexadecimal import (
    add_0x_prefix,
    decode_hex,
    encode_hex,
    is_hexstr,
    remove_0x_prefix,
)
from .types import is_boolean, is_integer, is_string

T = TypeVar("T")


@validate_conversion_arguments
def to_hex(
    primitive: Primitives = None, hexstr: HexStr = None, text: str = None
) -> HexStr:
    """
    Auto converts any supported value into its hex representation.
    Trims leading zeros, as defined in:
    https://github.com/ethereum/wiki/wiki/JSON-RPC#hex-value-encoding
    """
    if hexstr is not None:
        return add_0x_prefix(HexStr(hexstr.lower()))

    if text is not None:
        return encode_hex(text.encode("utf-8"))

    if is_boolean(primitive):
        return HexStr("0x1") if primitive else HexStr("0x0")

    if isinstance(primitive, (bytes, bytearray)):
        return encode_hex(primitive)
    elif is_string(primitive):
        raise TypeError(
            "Unsupported type: The primitive argument must be one of: bytes,"
            "bytearray, int or bool and not str"
        )

    if is_integer(primitive):
        return HexStr(hex(primitive))

    raise TypeError(
        "Unsupported type: '{0}'.  Must be one of: bool, str, bytes, bytearray"
        "or int.".format(repr(type(primitive)))
    )


@validate_conversion_arguments
def to_int(
    primitive: Primitives = None, hexstr: HexStr = None, text: str = None
) -> int:
    """
    Converts value to its integer representation.
    Values are converted this way:

     * primitive:

       * bytes, bytearrays: big-endian integer
       * bool: True => 1, False => 0
     * hexstr: interpret hex as integer
     * text: interpret as string of digits, like '12' => 12
    """
    if hexstr is not None:
        return int(hexstr, 16)
    elif text is not None:
        return int(text)
    elif isinstance(primitive, (bytes, bytearray)):
        return big_endian_to_int(primitive)
    elif isinstance(primitive, str):
        raise TypeError("Pass in strings with keyword hexstr or text")
    elif isinstance(primitive, (int, bool)):
        return int(primitive)
    else:
        raise TypeError(
            "Invalid type.  Expected one of int/bool/str/bytes/bytearray.  Got "
            "{0}".format(type(primitive))
        )


@validate_conversion_arguments
def to_bytes(
    primitive: Primitives = None, hexstr: HexStr = None, text: str = None
) -> bytes:
    if is_boolean(primitive):
        return b"\x01" if primitive else b"\x00"
    elif isinstance(primitive, bytearray):
        return bytes(primitive)
    elif isinstance(primitive, bytes):
        return primitive
    elif is_integer(primitive):
        return to_bytes(hexstr=to_hex(primitive))
    elif hexstr is not None:
        if len(hexstr) % 2:
            # type check ignored here because casting an Optional arg to str is not possible
            hexstr = "0x0" + remove_0x_prefix(hexstr)  # type: ignore
        return decode_hex(hexstr)
    elif text is not None:
        return text.encode("utf-8")
    raise TypeError(
        "expected a bool, int, byte or bytearray in first arg, or keyword of hexstr or text"
    )


@validate_conversion_arguments
def to_text(
    primitive: Primitives = None, hexstr: HexStr = None, text: str = None
) -> str:
    if hexstr is not None:
        return to_bytes(hexstr=hexstr).decode("utf-8")
    elif text is not None:
        return text
    elif isinstance(primitive, str):
        return to_text(hexstr=primitive)
    elif isinstance(primitive, (bytes, bytearray)):
        return primitive.decode("utf-8")
    elif is_integer(primitive):
        byte_encoding = int_to_big_endian(primitive)
        return to_text(byte_encoding)
    raise TypeError("Expected an int, bytes, bytearray or hexstr.")


def text_if_str(
    to_type: Callable[..., T], text_or_primitive: Union[bytes, int, str]
) -> T:
    """
    Convert to a type, assuming that strings can be only unicode text (not a hexstr)

    :param to_type function: takes the arguments (primitive, hexstr=hexstr, text=text),
        eg~ to_bytes, to_text, to_hex, to_int, etc
    :param text_or_primitive bytes, str, int: value to convert
    """
    if isinstance(text_or_primitive, str):
        return to_type(text=text_or_primitive)
    else:
        return to_type(text_or_primitive)


def hexstr_if_str(
    to_type: Callable[..., T], hexstr_or_primitive: Union[bytes, int, str]
) -> T:
    """
    Convert to a type, assuming that strings can be only hexstr (not unicode text)

    :param to_type function: takes the arguments (primitive, hexstr=hexstr, text=text),
        eg~ to_bytes, to_text, to_hex, to_int, etc
    :param hexstr_or_primitive bytes, str, int: value to convert
    """
    if isinstance(hexstr_or_primitive, str):
        if remove_0x_prefix(HexStr(hexstr_or_primitive)) and not is_hexstr(
            hexstr_or_primitive
        ):
            raise ValueError(
                "when sending a str, it must be a hex string. Got: {0!r}".format(
                    hexstr_or_primitive
                )
            )
        return to_type(hexstr=hexstr_or_primitive)
    else:
        return to_type(hexstr_or_primitive)



import hashlib

from platone.packages.eth_utils import (
    is_boolean,
    is_bytes,
    is_dict,
    is_list_like,
    is_null,
    is_number,
    is_text,
    to_bytes,
)

from .six import (
    Generator,
)


def generate_cache_key(value):
    """
    Generates a cache key for the *args and **kwargs
    """
    if is_bytes(value):
        return hashlib.md5(value).hexdigest()
    elif is_text(value):
        return generate_cache_key(to_bytes(text=value))
    elif is_boolean(value) or is_null(value) or is_number(value):
        return generate_cache_key(repr(value))
    elif is_dict(value):
        return generate_cache_key((
            (key, value[key])
            for key
            in sorted(value.keys())
        ))
    elif is_list_like(value) or isinstance(value, Generator):
        return generate_cache_key("".join((
            generate_cache_key(item)
            for item
            in value
        )))
    else:
        raise TypeError("Cannot generate cache key for value {0} of type {1}".format(
            value,
            type(value),
        ))
